print("import: 'pymc3'")
import pymc3

print("import: 'pymc3.backends'")
import pymc3.backends

print("import: 'pymc3.distributions'")
import pymc3.distributions

print("import: 'pymc3.glm'")
import pymc3.glm

print("import: 'pymc3.gp'")
import pymc3.gp

print("import: 'pymc3.ode'")
import pymc3.ode

print("import: 'pymc3.plots'")
import pymc3.plots

print("import: 'pymc3.smc'")
import pymc3.smc

print("import: 'pymc3.stats'")
import pymc3.stats

print("import: 'pymc3.step_methods'")
import pymc3.step_methods

print("import: 'pymc3.tests'")
import pymc3.tests

print("import: 'pymc3.tuning'")
import pymc3.tuning

print("import: 'pymc3.variational'")
import pymc3.variational

